@@ -91,7 +91,7 @@ DEFINE_MODEL(usbGraphire4,	"USB Graphire4",	4);
 DEFINE_MODEL(usbBamboo,		"USB Bamboo",		4);
 DEFINE_MODEL(usbBamboo1,	"USB Bamboo1",		4);
 DEFINE_MODEL(usbBambooFun,	"USB BambooFun",	4);
-DEFINE_MODEL(usbCintiq,		"USB PL/Cintiq",	4);
+DEFINE_MODEL(usbCintiq,		"USB PL/Cintiq/DTU",	4);
 DEFINE_MODEL(usbCintiqPartner,	"USB CintiqPartner",	4);
 DEFINE_MODEL(usbIntuos,		"USB Intuos1",		5);
 DEFINE_MODEL(usbIntuos2,	"USB Intuos2",		5);
@@ -167,7 +167,7 @@ static unsigned short padkey_codes [] = {
 /* Fixed mapped stylus and mouse buttons */
 
 #define WCM_USB_MAX_MOUSE_BUTTONS 5
-#define WCM_USB_MAX_STYLUS_BUTTONS 3
+#define WCM_USB_MAX_STYLUS_BUTTONS 4
 
 static unsigned short mouse_codes [] = {
 	BTN_LEFT, BTN_MIDDLE, BTN_RIGHT, BTN_BACK, BTN_FORWARD,
@@ -232,6 +232,8 @@ static struct WacomModelDesc
 	{ WACOM_VENDOR_ID, 0xC7, 100000, 100000, &usbCintiq,     "DTU1931"		},
 	{ WACOM_VENDOR_ID, 0xCE, 100000, 100000, &usbCintiq,     "DTU2231"		},
 	{ WACOM_VENDOR_ID, 0xF0, 100000, 100000, &usbCintiq,     "DTU1631"		},
+	{ WACOM_VENDOR_ID, 0x35a, 100000, 100000, &usbCintiq,    "DTH1152"              },
+	{ WACOM_VENDOR_ID, 0x368, 100000, 100000, &usbCintiq,    "DTH1152"              }, /* Touch */
 
 	{ WACOM_VENDOR_ID, 0x41, 100000, 100000, &usbIntuos2,    "Intuos2 4x5"		},
 	{ WACOM_VENDOR_ID, 0x42, 100000, 100000, &usbIntuos2,    "Intuos2 6x8"		},
@@ -287,6 +289,21 @@ static struct WacomModelDesc
 	{ WACOM_VENDOR_ID, 0x57, 200000, 200000, &usbCintiqV5,   "DTK2241"		},
 	{ WACOM_VENDOR_ID, 0x59, 200000, 200000, &usbCintiqV5,   "DTH2242"		},
 	{ WACOM_VENDOR_ID, 0x5B, 200000, 200000, &usbCintiqV5,   "Cintiq 22HDT"		},
+	{ WACOM_VENDOR_ID, 0x32B, 200000, 200000, &usbCintiqV5,  "Cintiq 27QHDT"	},
+	{ WACOM_VENDOR_ID, 0x32C, 200000, 200000, &usbCintiqV5,  "Cintiq 27QHDT"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x34A, 200000, 200000, &usbCintiqV5,  "Mobilestudio Pro 13"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x34B, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 16"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x34D, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 13"	},
+	{ WACOM_VENDOR_ID, 0x34E, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 16"	},
+	{ WACOM_VENDOR_ID, 0x34F, 200000, 200000, &usbCintiqV5,  "Cintiq 13 FHD Pro"	},
+	{ WACOM_VENDOR_ID, 0x350, 200000, 200000, &usbCintiqV5,  "Cintiq 16 UHD Pro"	},
+	{ WACOM_VENDOR_ID, 0x351, 200000, 200000, &usbCintiqV5,  "Cintiq Pro 24"	},
+	{ WACOM_VENDOR_ID, 0x352, 200000, 200000, &usbCintiqV5,  "Cintiq Pro 32"	},
+	{ WACOM_VENDOR_ID, 0x353, 200000, 200000, &usbCintiqV5,  "Cintiq 13 FHD Pro"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x354, 200000, 200000, &usbCintiqV5,  "Cintiq 16 UHD Pro"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x355, 200000, 200000, &usbCintiqV5,  "Cintiq Pro 24"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x356, 200000, 200000, &usbCintiqV5,  "Cintiq Pro 32"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x37C, 200000, 200000, &usbCintiqV5,  "Cintiq Pro 24"	}, /* Pen-only model */
 
 	{ WACOM_VENDOR_ID, 0x90, 100000, 100000, &usbTabletPC,   "TabletPC 0x90"	},
 	{ WACOM_VENDOR_ID, 0x93, 100000, 100000, &usbTabletPC,   "TabletPC 0x93"	},
@@ -401,6 +418,14 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *versi
 
 	DBG(1, priv, "initializing USB tablet\n");
 
+	/* fetch vendor, product, and model name */
+	if (ioctl(pInfo->fd, EVIOCGID, &sID) == -1 ||
+	    ioctl(pInfo->fd, EVIOCGNAME(id_len), id) == -1) {
+		xf86Msg(X_ERROR, "%s: failed to ioctl ID or name.\n",
+					pInfo->name);
+		return !Success;
+	}
+
 	if (!common->private &&
 	    !(common->private = calloc(1, sizeof(wcmUSBData))))
 	{
@@ -412,10 +437,6 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *versi
 	usbdata = common->private;
 	*version = 0.0;
 
-	/* fetch vendor, product, and model name */
-	ioctl(pInfo->fd, EVIOCGID, &sID);
-	ioctl(pInfo->fd, EVIOCGNAME(id_len), id);
-
 	for (i = 0; i < ARRAY_SIZE(WacomModelDesc); i++)
 	{
 		if (sID.vendor == WacomModelDesc[i].vendor_id &&
@@ -459,7 +480,7 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *versi
 	/* nbuttons tracks maximum buttons on all tools (stylus/mouse).
 	 *
 	 * Mouse support left, middle, right, side, and extra side button.
-	 * Stylus support tip and 2 stlyus buttons.
+	 * Stylus support tip and 3 stylus buttons.
 	 */
 	if (ISBITSET (common->wcmKeys, BTN_TOOL_MOUSE))
 		usbdata->nbuttons = WCM_USB_MAX_MOUSE_BUTTONS;
@@ -549,7 +570,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 		/* may be an expresskey only interface */
 		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
 		    ISBITSET(common->wcmKeys, BTN_0))
-			return Success;
+			goto pad_init;
 
 		xf86Msg(X_ERROR, "%s: no abs bits.\n", pInfo->name);
 		return !Success;
@@ -568,7 +589,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 		/* may be a PAD only interface */
 		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
 		    ISBITSET(common->wcmKeys, BTN_0))
-			return Success;
+			goto pad_init;
 
 		xf86Msg(X_ERROR, "%s: unable to ioctl xmax value.\n", pInfo->name);
 		return !Success;
@@ -648,6 +669,16 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 			common->wcmMaxStripX = absinfo.maximum;
 	}
 
+	/* max touchring value for standalone pad tools */
+	common->wcmMinRing = 0;
+	common->wcmMaxRing = 71;
+	if (!ISBITSET(ev,EV_MSC) && ISBITSET(abs, ABS_WHEEL) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_WHEEL), &absinfo))
+	{
+		common->wcmMinRing = absinfo.minimum;
+		common->wcmMaxRing = absinfo.maximum;
+	}
+
 	/* X tilt range */
 	if (ISBITSET(abs, ABS_TILT_X) &&
 			!ioctl(pInfo->fd, EVIOCGABS(ABS_TILT_X), &absinfo))
@@ -774,7 +805,8 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 		memset(sw, 0, sizeof(sw));
 
-		ioctl(pInfo->fd, EVIOCGSW(sizeof(sw)), sw);
+		if (ioctl(pInfo->fd, EVIOCGSW(sizeof(sw)), sw) < 0)
+			xf86Msg(X_ERROR, "%s: unable to ioctl sw state.\n", pInfo->name);
 
 		if (ISBITSET(sw, SW_MUTE_DEVICE))
 			common->wcmHWTouchSwitchState = 0;
@@ -782,6 +814,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 			common->wcmHWTouchSwitchState = 1;
 	}
 
+pad_init:
 	usbWcmInitPadState(pInfo);
 
 	return Success;
@@ -809,11 +842,13 @@ static int usbParse(InputInfoPtr pInfo, const unsigned char* data, int len)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
+	struct input_event event;
 
 	if (len < sizeof(struct input_event))
 		return 0;
 
-	usbParseEvent(pInfo, (const struct input_event*)data);
+	memcpy(&event, data, sizeof(event));
+	usbParseEvent(pInfo, &event);
 	return common->wcmPktLength;
 }
 
@@ -1259,6 +1294,8 @@ static void usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 			if (event->value >= 0) {
 				int serial = event->value + 1;
 				private->wcmMTChannel = usbChooseChannel(common, TOUCH_ID, serial);
+				if (private->wcmMTChannel < 0)
+					return;
 				ds = &common->wcmChannel[private->wcmMTChannel].work;
 				ds->serial_num = serial;
 			}
@@ -1424,6 +1461,10 @@ static void usbParseKeyEvent(WacomCommonPtr common,
 			ds->buttons = mod_buttons(ds->buttons, 2, event->value);
 			break;
 
+		case BTN_STYLUS3:
+			ds->buttons = mod_buttons(ds->buttons, 3, event->value);
+			break;
+
 		default:
 			change = 0;
 	}
@@ -1791,10 +1832,10 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	}
 
 	/* verify we have minimal data when entering prox */
-	if (ds->proximity && !dslast.proximity) {
+	if (ds->proximity && !dslast.proximity && ds->device_type != PAD_ID) {
 		struct input_absinfo absinfo;
 
-		if (!ds->x && !IsPad(priv)) {
+		if (!ds->x) {
 			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
 			{
 				DBG(-1, common, "unable to ioctl current x value.\n");
@@ -1802,7 +1843,7 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 			}
 			ds->x = absinfo.value;
 		}
-		if (!ds->y && !IsPad(priv)) {
+		if (!ds->y) {
 			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_Y), &absinfo) < 0)
 			{
 				DBG(-1, common, "unable to ioctl current x value.\n");
